module coeffs
!--------------------------------------------------------------------------------------
! Coefficient and boundary condition functions that define the general 2D linear PDE:
!
!     au_xx + bu_yy + cu_x + du_y + eu = f,         (1)
!     subject to u = BC on the boundaries           (2)
! 
! The solution u(x,y) is actually given by the BC function. I started from that,
! made up a-e, and then computed f to equal the left hand side of (1). See the
! Matlab script cool.m for the actual computations of the functions.
!
! Another note: although I started converting all of the constants like 32 to
! 32.0_real8, it became onerous. Since most of the constants are powers of two,
! I just made them 32.0, etc.
!-----------------
! Randall Bramley
! Department of Computer Science
! Indiana University, Bloomington
!--------------------------------------------------------------------------------------
    use kindlytypes, only: real8
    real(kind=real8), parameter :: pi = 3.1415926535897924_real8

contains 
    real(kind=real8) function afun(x,y) 
        real(kind=real8) :: x, y
        afun =  exp(x - y)
        return
    end function afun

    real(kind=real8) function  bfun(x,y) 
        real(kind=real8) :: x, y
        bfun = asinh(x + y + 1.0_real8)
        return
    end function bfun

    real(kind=real8) function  cfun(x,y) 
        real(kind=real8) :: x, y
        cfun = pi*y
        return
    end function cfun

    real(kind=real8) function  dfun(x,y) 
        real(kind=real8) :: x, y
        dfun = x*y
        return
    end function dfun

    real(kind=real8) function  efun(x,y) 
        real(kind=real8) :: x, y
        efun = 1.0_real8
        return
    end function efun

    real(kind=real8) function  ffun(x,y) 
        real(kind=real8) :: x, y, s
        real(kind=real8), parameter :: half = 0.5_real8, two = 2.0_real8

        s = 16*pi**2*(x - half)**2 + 4*pi**2*(y - half)**2 + two
        ffun = sin(4*pi*(x - half) + two*pi*(y - half)) - cos(4*pi*(x - half) &
            - two*pi*(y - half)) + exp(x - y)* &
            (16.0*pi**2*cos(4.0*pi*(x - half) - two*pi*(y - half)) - &
            (32.0*pi**2)/s**2 - 16*pi**2*sin(4.0*pi*(x - half) + two*pi*(y - half)) &
            + (512.0*pi**4*(two*x - 1.0)**2)/s**3) + asinh(x + y + 1.0)* &
            (4.0*pi**2*cos(4.0*pi*(x - half) - two*pi*(y - half)) - &
            (8.0*pi**2)/s**2 - 4.0*pi**2*sin(4.0*pi*(x - half) + two*pi*(y - half)) &
            + (32.0*pi**4*(two*y - 1.0)**2)/s**3) + 1.0/s + &
            y*pi*(4.0*pi*cos(4.0*pi*(x - half) + two*pi*(y - half)) + &
            4.0*pi*sin(4.0*pi*(x - half) - two*pi*(y - half)) - &
            (16.0*pi**2*(two*x - 1.0))/s**2) - x*y*(two*pi*sin(4.0*pi*(x - half) - &
            two*pi*(y - half)) - two*pi*cos(4.0*pi*(x - half) + &
            two*pi*(y - half)) + (4.0*pi**2*(two*y - 1.0))/s**2)
        return
    end function ffun

    real(kind=real8) function BC(x,y) 
        real(kind=real8) :: x, y
        real(kind=real8), parameter :: half = 0.5_real8, two = 2.0_real8
        BC = sin(4.0*pi*(x - half) + &
            two*pi*(y - half)) - cos(4.0*pi*(x - half) - &
            two*pi*(y - half)) + 1.0/(16.0*pi**2*(x - half)**2 + &
            4.0*pi**2*(y - half)**2 + two)
        return
    end function BC
        
end module coeffs
